﻿using System;
using System.Windows.Media;

namespace Fantasia.Imaging.Silverlight.Utilities
{
	public static class ColorHelper
	{
		/// <summary>
		/// Blends a front color on a back color, assuming the backcolor 
		/// </summary>
		/// <param name="back">premultiplied back color</param>
		/// <param name="frontColor">premultiplied front color</param>
		/// <returns></returns>
		public static Color BlendPremuls(Color back, uint frontColor)
		{
			Color result;
			uint sa = (frontColor >> 24);
			uint sr = ((frontColor >> 16) & 0xff);
			uint sg = ((frontColor >> 8) & 0xff);
			uint sb = ((frontColor) & 0xff);
			if (sa == 255) // all source
			{
				result = Color.FromArgb((byte)sa, (byte)sr, (byte)sg, (byte)sb);
				return result;
			}

			uint da = back.A;
			uint dr = back.R;
			uint dg = back.G;
			uint db = back.B;

			result = Color.FromArgb((byte)(sa + (((da * (255 - sa)) * 0x8081) >> 23)),
				(byte)(sr + (((dr * (255 - sa)) * 0x8081) >> 23)),
				(byte)(sg + (((dg * (255 - sa)) * 0x8081) >> 23)),
				(byte)(sb + (((db * (255 - sa)) * 0x8081) >> 23)));
			return result;
		}

		/// <summary>
		/// Retrieves an int representation of the color object
		/// </summary>
		/// <param name="c"></param>
		/// <returns></returns>
		public static int IntFromColor(Color c)
		{
			return (c.A << 24) | (c.R << 16) | (c.G << 8) | c.B;
		}
		
		/// <summary>
		/// Retrieves an int representation of the color object
		/// </summary>
		/// <param name="c"></param>
		/// <returns></returns>
		public static uint UIntFromColor(Color c)
		{
			return unchecked((uint)((c.A << 24) | (c.R << 16) | (c.G << 8) | c.B));
		}

		/// <summary>
		/// Retrieves an int representation of the color object
		/// </summary>
		/// <param name="c"></param>
		/// <returns></returns>
		public static Color ColorFromUint(uint c)
		{
			return Color.FromArgb((byte)(c >> 24), (byte)(c >> 16), (byte)(c >> 8), (byte)c);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="hash"></param>
		/// <param name="darkness">0..1 (0 is darkest)</param>
		/// <returns></returns>
		public static Color FriendlyColorFromHash(int hash, double darkness)
		{
			if (hash < 0) hash = -hash;
			return HsvToRgb(hash % 360, 1, darkness);
		}

		/// <summary>
		/// Amount of clamping in % to do on each side of the picker control
		/// Only applied to the color section (gray section is not clamped)
		/// </summary>
		public const double QUICK_PICKER_CLAMP = 0.07;
		public const int QUICK_PICKER_COLOR_SECTION_WIDTH = 360;
		public const int QUICK_PICKER_GRAY_SECTION_WIDTH = 40;
		public const int QUICK_PICKER_HEIGHT = 60;
		public const int QUICK_PICKER_WIDTH = QUICK_PICKER_COLOR_SECTION_WIDTH + QUICK_PICKER_GRAY_SECTION_WIDTH;

		/// <summary>
		/// Retrieves line color picker coords for a given color value
		/// </summary>
		public static void GetLinePickerCoordinatesFromColor(int r, int g, int b, out int x, out int y)
		{
			float h, s, v;
			ColorHelper.ConvertRgbToHsv(r / 255.0f, g / 255.0f, b / 255.0f, out h, out s, out v);

			if (s == 0) // grayscale
			{
				x = QUICK_PICKER_COLOR_SECTION_WIDTH + QUICK_PICKER_GRAY_SECTION_WIDTH / 2;
				y = (int)((1 - v) * QUICK_PICKER_HEIGHT);
				if (y < 0) y = 0;
			}
			else
			{
				x = (int)h;
				double yn;
				// below if's come with color loss: we assume either saturation or value is 1
				if (v > s) // y < 0.5: top half (color to white)
				{
					yn = s * 0.5;
				}
				else // bottom half: color to black
				{
					yn = 1 - v * 0.5;
				}

				yn = (yn - QUICK_PICKER_CLAMP) / (1 - QUICK_PICKER_CLAMP * 2);
				y = (int)(yn * QUICK_PICKER_HEIGHT);
				if (y < 0) y = 0;
				if (y >= QUICK_PICKER_HEIGHT) y = QUICK_PICKER_HEIGHT - 1;
			}
		}

		const int COLOR_PICKER_WIDTH = 360;
		const int COLOR_PICKER_HEIGHT = 360;

		/// <summary>
		/// Retrieves color wheel coords for a given color value
		/// </summary>
		/// <param name="r"></param>
		/// <param name="g"></param>
		/// <param name="b"></param>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <param name="z"></param>
		public static void GetWheelCoordinatesFromColor(int r, int g, int b, out int x, out int y, out int z)
		{
			// light = v == 1, s = pos to white
			// dark = s = 1.0, v = pos to black
			// grayscale = s== 0

			const int CX = COLOR_PICKER_WIDTH / 2;
			const int CY = COLOR_PICKER_HEIGHT / 2;
			const int RADIUS = (int)(0.6 * CX); // 360 is the width of the color picker, 0.6 is the point of falloff

			if ((r == 0) && (g == 0) && (b == 0)) // pitch black: special case
			{
				x = CX;
				y = CY - RADIUS - 4;
				z = 127;
				return;
			}

			// h is the angle on the color picker, RADIUS is the radius from white to full saturation
			float h, s, v;
			ColorHelper.ConvertRgbToHsv(r / 255.0f, g / 255.0f, b / 255.0f, out h, out s, out v);
			h = h * Constants.DEG_TO_RADf;

			double xPos = Math.Sin(h) * RADIUS * s;
			double yPos = Math.Cos(h) * RADIUS * s;

			x = -((int)xPos) + CX;
			y = -((int)yPos) + CY;
			z = 255 - ((int)(v * 128));
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="h">in the range -360..360</param>
		/// <param name="s">0..1 (white to color)</param>
		/// <param name="v">0..1 (black to color)</param>
		/// <returns></returns>
		public static Color HsvToRgb(double h, double s, double v)
		{
			return HsvToRgbTransferAlpha(255, h, s, v);
		}

		public static void HueToColor(float h, out int r, out int g, out int b)
		{
			if (h < 0) h += 360;

			float hf = h / 60.0f;
			int i = (int)Math.Floor(hf);
			float f = hf - i;

			switch (i)
			{
				case 0:
					r = 255;
					g = (int)(f * 255);
					b = 0;
					break;

				case 1:
					r = (int)((1 - f) * 255);
					g = 255;
					b = 0;
					break;

				case 2:
					r = 0;
					g = 255;
					b = (int)(f * 255);
					break;

				case 3:
					r = 0;
					g = (int)((1 - f) * 255);
					b = 255;
					break;

				case 4:
					r = (int)(f * 255);
					g = 0;
					b = 1;
					break;

				case 5:
					r = 255;
					g = 0;
					b = (int)((1 - f) * 255);
					break;

				case 6:
					r = 255;
					g = (int)(f * 255);
					b = 0;
					break;

				default:
					r = g = b = 255;
					break;
			}

			//r = Clamp((int)(R * 255.0));
			//g = Clamp((int)(G * 255.0));
			//b = Clamp((int)(B * 255.0));
		}

		public static Color HsvToRgbTransferAlpha(byte a, double h, double s, double v)
		{
			if (h < 0) h += 360;
			Color result = new Color();
			result.A = a;

			double R, G, B;
			if (v <= 0)
			{
				R = G = B = 0;
			}
			else if (s <= 0)
			{
				R = G = B = v;
			}
			else
			{
				double hf = h / 60.0;
				int i = (int)Math.Floor(hf);
				double f = hf - i;
				double pv = v * (1 - s);
				double qv = v * (1 - s * f);
				double tv = v * (1 - s * (1 - f));

				switch (i)
				{
					case 0:
						R = v;
						G = tv;
						B = pv;
						break;

					case 1:
						R = qv;
						G = v;
						B = pv;
						break;

					case 2:
						R = pv;
						G = v;
						B = tv;
						break;

					case 3:
						R = pv;
						G = qv;
						B = v;
						break;

					case 4:
						R = tv;
						G = pv;
						B = v;
						break;

					case 5:
						R = v;
						G = pv;
						B = qv;
						break;

					case 6:
						R = v;
						G = tv;
						B = pv;
						break;

					default:
						R = G = B = v;
						break;
				}
			}

			result.R = Clamp((int)(R * 255.0));
			result.G = Clamp((int)(G * 255.0));
			result.B = Clamp((int)(B * 255.0));

			return result;
		}

		/// <summary>
		/// Clamp a value to 0-255
		/// </summary>
		private static byte Clamp(int i)
		{
			if (i < 0) return 0;
			if (i > 255) return 255;
			return (byte)i;
		}

		private const byte MIN = 0;
		private const byte MAX = 255;

		public static Color GetColorFromPosition(int position)
		{
			byte mod = (byte)(position % MAX);
			byte diff = (byte)(MAX - mod);
			byte alpha = 255;

			switch (position / MAX)
			{
				case 0: return Color.FromArgb(alpha, MAX, mod, MIN);
				case 1: return Color.FromArgb(alpha, diff, MAX, MIN);
				case 2: return Color.FromArgb(alpha, MIN, MAX, mod);
				case 3: return Color.FromArgb(alpha, MIN, diff, MAX);
				case 4: return Color.FromArgb(alpha, mod, MIN, MAX);
				case 5: return Color.FromArgb(alpha, MAX, MIN, diff);
				default: return Colors.Black;
			}
		}

		public static string GetHexCode(Color c)
		{
			return string.Format("#{0}{1}{2}{3}",
				c.A.ToString("X2"),
				c.R.ToString("X2"),
				c.G.ToString("X2"),
				c.B.ToString("X2"));
		}

		/// <summary>
		/// Converts RGB color to HSV color
		/// </summary>
		/// <param name="r">Red [0..1]</param>
		/// <param name="g">Green [0..1]</param>
		/// <param name="b">Blur [0..1]</param>
		/// <param name="h">Hue [0..360]</param>
		/// <param name="s">Saturation [0..1]</param>
		/// <param name="v">Value [0..1]</param>
		public static void ConvertRgbToHsv(float r, float g, float b, out float h, out float s, out float v)
		{
			float colorMax = Math.Max(Math.Max(r, g), b);

			v = colorMax;
			if (v == 0)
			{
				h = 0;
				s = 0;
				return;
			}

			// normalize
			r /= v;
			g /= v;
			b /= v;

			float colorMin = Math.Min(Math.Min(r, g), b);
			colorMax = Math.Max(Math.Max(r, g), b);

			s = colorMax - colorMin;
			if (s == 0)
			{
				h = 0;
				return;
			}

			// normalize saturation
			r = (r - colorMin) / s;
			g = (g - colorMin) / s;
			b = (b - colorMin) / s;
			colorMax = Math.Max(Math.Max(r, g), b);

			// calculate hue
			if (colorMax == r)
			{
				h = 60 * (g - b);
				if (h < 0.0)
				{
					h += 360;
				}
			}
			else if (colorMax == g)
			{
				h = 120 + 60 * (b - r);
			}
			else // colorMax == b
			{
				h = 240 + 60 * (r - g);
			}

		}

		// Algorithm ported from: http://www.colorjack.com/software/dhtml+color+picker.html
		/// <summary>
		/// Converts HSV to RGB color
		/// </summary>
		/// <param name="h">Hue [0..360]</param>
		/// <param name="s">Saturation [0..1]</param>
		/// <param name="v">Value [0..1]</param>
		/// <returns>RGB color</returns>
		public static Color ConvertHsvToRgb(float h, float s, float v)
		{
			h = h / 360;
			if (s > 0)
			{
				if (h >= 1)
					h = 0;
				h = 6 * h;
				int hueFloor = (int)Math.Floor(h);
				byte a = (byte)Math.Round(MAX * v * (1.0 - s));
				byte b = (byte)Math.Round(MAX * v * (1.0 - (s * (h - hueFloor))));
				byte c = (byte)Math.Round(MAX * v * (1.0 - (s * (1.0 - (h - hueFloor)))));
				byte d = (byte)Math.Round(MAX * v);

				switch (hueFloor)
				{
					case 0: return Color.FromArgb(MAX, d, c, a);
					case 1: return Color.FromArgb(MAX, b, d, a);
					case 2: return Color.FromArgb(MAX, a, d, c);
					case 3: return Color.FromArgb(MAX, a, b, d);
					case 4: return Color.FromArgb(MAX, c, a, d);
					case 5: return Color.FromArgb(MAX, d, a, b);
					default: return Color.FromArgb(0, 0, 0, 0);
				}
			}
			else
			{
				byte d = (byte)(v * MAX);
				return Color.FromArgb(255, d, d, d);
			}
		}

		/// <summary>
		/// Derives a new color from existing color by shifting values in HSV space
		/// Can be used to create pleasant color mixes (for example, for Fairy Lights brush)
		/// </summary>
		/// <param name="color">Primary color</param>
		/// <param name="hueShiftDegrees">Hue shift in degrees (0..360)</param>
		/// <param name="saturationShiftPercentPoints">Saturation shift in percent points (e.g. 0.02 will increase saturation by 2 points, range 0..1)</param>
		/// <param name="valueShiftPercentPoints">Value shift in percent points</param>
		/// <returns>Derived color after shifting the primary color in HSV space</returns>
		public static Color DeriveColor(Color color, float hueShiftDegrees, float saturationShiftPercentPoints, float valueShiftPercentPoints)
		{
			float h, s, v;
			ColorHelper.ConvertRgbToHsv(color.R / 255.0f, color.G / 255.0f, color.B / 255.0f, out h, out s, out v);
			h += hueShiftDegrees;
			s += saturationShiftPercentPoints;
			v += valueShiftPercentPoints;
			
			if (h < 0) h += 360;
			else if (h > 360) h -= 360;

			if (s < 0) s = 0;
			else if (s > 1) s = 1;
	
			if (v < 0) v = 0;
			else if (v > 1) v = 1;

			return ColorHelper.ConvertHsvToRgb(h, s, v);
		}

		#region Color Matrix

		/// <summary>
		/// Multiply 2 matrices
		/// </summary>
		/// <param name="a"></param>
		/// <param name="b"></param>
		/// <param name="c"></param>
		public static void MatrixMul(float[] a, float[] b, float[] c)
		{
			float[] temp = new float[16];

			for (int y = 0; y < 4; y++)
				for (int x = 0; x < 4; x++)
				{
					temp[(y << 2) + x] = b[(y << 2) + 0] * a[(0 << 2) + x]
							   + b[(y << 2) + 1] * a[(1 << 2) + x]
							   + b[(y << 2) + 2] * a[(2 << 2) + x]
							   + b[(y << 2) + 3] * a[(3 << 2) + x];
				}

			Array.Copy(temp, 0, c, 0, 16);
		}

		/// <summary>
		/// Return an identity matrix
		/// </summary>
		/// <param name="m"></param>
		public static float[] MatrixIdentity()
		{
			float[] m = new float[16];
			m[(0 << 2) + 0] = 1;
			m[(1 << 2) + 1] = 1;
			m[(2 << 2) + 2] = 1;
			m[(3 << 2) + 3] = 1;

			return m;
		}

		/// <summary>
		/// Transform a 3D point using a matrix
		/// </summary>
		/// <param name="matrix"></param>
		/// <param name="x"></param>
		/// <param name="y"></param>
		/// <param name="z"></param>
		/// <param name="rx"></param>
		/// <param name="ry"></param>
		/// <param name="rz"></param>
		public static void MatrixTransform(float[] matrix, float x, float y, float z, out float rx, out float ry, out float rz)
		{
			rx = x * matrix[(0 << 2) + 0] + y * matrix[(1 << 2) + 0] + z * matrix[(2 << 2) + 0] + matrix[(3 << 2) + 0];
			ry = x * matrix[(0 << 2) + 1] + y * matrix[(1 << 2) + 1] + z * matrix[(2 << 2) + 1] + matrix[(3 << 2) + 1];
			rz = x * matrix[(0 << 2) + 2] + y * matrix[(1 << 2) + 2] + z * matrix[(2 << 2) + 2] + matrix[(3 << 2) + 2];
		}

		const float SQRT_2 = 1.4142135623730950488016887242097f;
		const float SQRT_3 = 1.7320508075688772935274463415059f;
		const float RLUM = 0.3086f;
		const float GLUM = 0.6094f;
		const float BLUM = 0.0820f;

		/// <summary>
		/// Make a matrix that rotates hue by an angle
		/// </summary>
		/// <param name="mat"></param>
		/// <param name="rot"></param>
		public static void MatrixRotateHue(float[] mat, float rot)
		{
			float[] mmat = MatrixIdentity();
			float mag;
			float lx, ly, lz;
			float xrs, xrc;
			float yrs, yrc;
			float zrs, zrc;
			float zsx, zsy;

			// rotate the grey vector into positive Z
			mag = SQRT_2;
			xrs = 1.0f / mag;
			xrc = 1.0f / mag;
			MatrixRotateX(mmat, xrs, xrc);
			mag = SQRT_3;
			yrs = -1.0f / mag;
			yrc = SQRT_2 / mag;
			MatrixRotateY(mmat, yrs, yrc);

			// shear the space to make the luminance plane horizontal
			MatrixTransform(mmat, RLUM, GLUM, BLUM, out lx, out ly, out lz);
			zsx = lx / lz;
			zsy = ly / lz;
			MatrixShearZ(mmat, zsx, zsy);

			// rotate hue 
			zrs = (float)Math.Sin(rot * Constants.PIf / 180.0f);
			zrc = (float)Math.Cos(rot * Constants.PIf / 180.0f);
			MatrixRotateZ(mmat, zrs, zrc);

			// unshear the space to put the luminance plane back 
			MatrixShearZ(mmat, -zsx, -zsy);

			// rotate the grey vector back into place
			MatrixRotateY(mmat, -yrs, yrc);
			MatrixRotateX(mmat, -xrs, xrc);

			MatrixMul(mmat, mat, mat);
		}

		/// <summary>
		/// Rotate around X (red) axis
		/// </summary>
		/// <param name="mat"></param>
		/// <param name="rs"></param>
		/// <param name="rc"></param>
		public static void MatrixRotateX(float[] mat, float rs, float rc)
		{
			float[] mmat = new float[16];

			mmat[(0 << 2) + 0] = 1.0f;
			mmat[(0 << 2) + 1] = 0.0f;
			mmat[(0 << 2) + 2] = 0.0f;
			mmat[(0 << 2) + 3] = 0.0f;

			mmat[(1 << 2) + 0] = 0.0f;
			mmat[(1 << 2) + 1] = rc;
			mmat[(1 << 2) + 2] = rs;
			mmat[(1 << 2) + 3] = 0.0f;

			mmat[(2 << 2) + 0] = 0.0f;
			mmat[(2 << 2) + 1] = -rs;
			mmat[(2 << 2) + 2] = rc;
			mmat[(2 << 2) + 3] = 0.0f;

			mmat[(3 << 2) + 0] = 0.0f;
			mmat[(3 << 2) + 1] = 0.0f;
			mmat[(3 << 2) + 2] = 0.0f;
			mmat[(3 << 2) + 3] = 1.0f;
			MatrixMul(mmat, mat, mat);
		}

		/// <summary>
		/// Rotate around y (green) axis
		/// </summary>
		/// <param name="mat"></param>
		/// <param name="rs"></param>
		/// <param name="rc"></param>
		public static void MatrixRotateY(float[] mat, float rs, float rc)
		{
			float[] mmat = new float[16];

			mmat[(0 << 2) + 0] = rc;
			mmat[(0 << 2) + 1] = 0.0f;
			mmat[(0 << 2) + 2] = -rs;
			mmat[(0 << 2) + 3] = 0.0f;

			mmat[(1 << 2) + 0] = 0.0f;
			mmat[(1 << 2) + 1] = 1.0f;
			mmat[(1 << 2) + 2] = 0.0f;
			mmat[(1 << 2) + 3] = 0.0f;

			mmat[(2 << 2) + 0] = rs;
			mmat[(2 << 2) + 1] = 0.0f;
			mmat[(2 << 2) + 2] = rc;
			mmat[(2 << 2) + 3] = 0.0f;

			mmat[(3 << 2) + 0] = 0.0f;
			mmat[(3 << 2) + 1] = 0.0f;
			mmat[(3 << 2) + 2] = 0.0f;
			mmat[(3 << 2) + 3] = 1.0f;
			MatrixMul(mmat, mat, mat);
		}

		/// <summary>
		/// Rotate around z (blue) axis
		/// </summary>
		/// <param name="mat"></param>
		/// <param name="rs"></param>
		/// <param name="rc"></param>
		public static void MatrixRotateZ(float[] mat, float rs, float rc)
		{
			float[] mmat = new float[16];

			mmat[(0 << 2) + 0] = rc;
			mmat[(0 << 2) + 1] = rs;
			mmat[(0 << 2) + 2] = 0.0f;
			mmat[(0 << 2) + 3] = 0.0f;

			mmat[(1 << 2) + 0] = -rs;
			mmat[(1 << 2) + 1] = rc;
			mmat[(1 << 2) + 2] = 0.0f;
			mmat[(1 << 2) + 3] = 0.0f;

			mmat[(2 << 2) + 0] = 0.0f;
			mmat[(2 << 2) + 1] = 0.0f;
			mmat[(2 << 2) + 2] = 1.0f;
			mmat[(2 << 2) + 3] = 0.0f;

			mmat[(3 << 2) + 0] = 0.0f;
			mmat[(3 << 2) + 1] = 0.0f;
			mmat[(3 << 2) + 2] = 0.0f;
			mmat[(3 << 2) + 3] = 1.0f;
			MatrixMul(mmat, mat, mat);
		}

		/// <summary>
		/// Shear Z using x and y
		/// </summary>
		/// <param name="mat"></param>
		/// <param name="dx"></param>
		/// <param name="dy"></param>
		public static void MatrixShearZ(float[] mat, float dx, float dy)
		{
			float[] mmat = new float[16];

			mmat[(0 << 2) + 0] = 1.0f;
			mmat[(0 << 2) + 1] = 0.0f;
			mmat[(0 << 2) + 2] = dx;
			mmat[(0 << 2) + 3] = 0.0f;

			mmat[(1 << 2) + 0] = 0.0f;
			mmat[(1 << 2) + 1] = 1.0f;
			mmat[(1 << 2) + 2] = dy;
			mmat[(1 << 2) + 3] = 0.0f;

			mmat[(2 << 2) + 0] = 0.0f;
			mmat[(2 << 2) + 1] = 0.0f;
			mmat[(2 << 2) + 2] = 1.0f;
			mmat[(2 << 2) + 3] = 0.0f;

			mmat[(3 << 2) + 0] = 0.0f;
			mmat[(3 << 2) + 1] = 0.0f;
			mmat[(3 << 2) + 2] = 0.0f;
			mmat[(3 << 2) + 3] = 1.0f;
			MatrixMul(mmat, mat, mat);
		}

		/// <summary>
		/// Make a saturation altering matrix
		/// </summary>
		/// <param name="mat"></param>
		/// <param name="sat"></param>
		public static void MatrixSaturation(float[] mat, float sat)
		{
			float[] mmat = MatrixIdentity();
			float a, b, c, d, e, f, g, h, i;
			float rwgt, gwgt, bwgt;

			rwgt = RLUM;
			gwgt = GLUM;
			bwgt = BLUM;

			a = (1.0f - sat) * rwgt + sat;
			b = (1.0f - sat) * rwgt;
			c = (1.0f - sat) * rwgt;
			d = (1.0f - sat) * gwgt;
			e = (1.0f - sat) * gwgt + sat;
			f = (1.0f - sat) * gwgt;
			g = (1.0f - sat) * bwgt;
			h = (1.0f - sat) * bwgt;
			i = (1.0f - sat) * bwgt + sat;
			mmat[(0 << 2) + 0] = a;
			mmat[(0 << 2) + 1] = b;
			mmat[(0 << 2) + 2] = c;
			mmat[(0 << 2) + 3] = 0.0f;

			mmat[(1 << 2) + 0] = d;
			mmat[(1 << 2) + 1] = e;
			mmat[(1 << 2) + 2] = f;
			mmat[(1 << 2) + 3] = 0.0f;

			mmat[(2 << 2) + 0] = g;
			mmat[(2 << 2) + 1] = h;
			mmat[(2 << 2) + 2] = i;
			mmat[(2 << 2) + 3] = 0.0f;

			mmat[(3 << 2) + 0] = 0.0f;
			mmat[(3 << 2) + 1] = 0.0f;
			mmat[(3 << 2) + 2] = 0.0f;
			mmat[(3 << 2) + 3] = 1.0f;
			MatrixMul(mmat, mat, mat);
		}
		#endregion
	}
}
